<?php


/*
 *
 *  toxA PHP Proxy
 *  Copyright (C) 2008  Thomas Urban, <info@toxa.de>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


// drop input data optionally available because of enabled register_globals
foreach ( $_REQUEST as $key => $value ) unset( $$key );


// read configuration
require( './.config.php' );


// set version
$version = '0.1alpha';


// start session
session_start();



$ERRNO = $ERRSTR = null;


try
{

	/*
	 * Get originally requested URI and arguments.
	 */

	// * all but first argument was queried by client originally
	$QUERY = $_SERVER['QUERY_STRING'];
	$QUERY = explode( '&', $QUERY );
	array_shift( $QUERY );
	$QUERY = implode( '&', $QUERY );

	// * first argument was inserted by mod_rewrite containing original URI
	$URI   = $_GET['__toxA'];

	// * fix name of index (supporting configuration-based customization)
	if ( $URI == 'index.html' )
		$URI = $indexName;



	/*
	 * Pre-analyse URLs set in configuration describing proxy and its target.
	 */

	// * analyse proxy address
	$PROXY = parse_url( $PROXY );

	// * analyse target
	$TARGET = parse_url( $TARGET );

	if ( !$PROXY || !$TARGET )
		throw new Exception( 'Invalid proxy configuration!' );

	// * extract target's port from its URL
	$PORT = ctype_digit( $TARGET['port'] ) ? intval( $TARGET['port'] ) : 80;


	// prepend target URI by target path
	$URI = $TARGET['path'] . '/' . $URI;

	if ( $QUERY )
		$URI .= '?' . $QUERY;





	/*
	 * Support local caching of files.
	 */

	$cacheDir  = dirname( __FILE__ ) . '/.cache';
	$cacheHash = false;

	if ( ( $_SERVER['REQUEST_METHOD'] == 'GET' ) && is_dir( $cacheDir ) )
	{

		if ( is_string( $useCache ) )
		{

			$pattern = trim( $useCache );
			$negate  = ( $pattern[0] == '!' );
			if ( $negate )
				$pattern = substr( $pattern, 1 );

			$m = preg_match( $useCache, $URI ) ^ $negate;

		}
		else
			$m = ( $useCache != false );

		if ( $m )
		{
			// cache is actually enabled on this request

			/*
			 * auto-clean cache from time to time
			 */

			$minTime = time() - $maxCacheAge;

			$info = @stat( "$cacheDir/.stamp" );
			if ( !$info || ( $info['mtime'] < $minTime ) )
			{

				$dir = @opendir( $cacheDir );
				if ( $dir )
				{

					while ( ( $name = readdir( $dir ) ) !== false )
						if ( $name[0] != '' )
						{

							$info = @stat( "$cacheDir/$name" );
							if ( $info && ( $info['mtime'] < $minTime ) )
								@unlink( "$cacheDir/$name" );

						}

					closedir( $dir );

				}

				@touch( "$cacheDir/.stamp" );

			}


			/*
			 * check cache for matching record
			 */

			$cacheHash = sha1( serialize( "$TARGET[host]/$URI" ) );

			$file      = "$cacheDir/$cacheHash";
			if ( file_exists( $file ) )
			{

				$stat = stat( $file );
				if ( time() - $stat['mtime'] < $maxCacheAge )
				{

					$record = @unserialize( file_get_contents( $file ) );
					if ( is_array( $record ) )
					{
						// got valid record from cache -> prefer that

						list( $state, $header, $data ) = $record;

						sendResponse( $state, $header, $data );

						exit( 0 );

					}
				}
			}
		}
	}





	/*
	 * Compile HTTP header for retreiving requested resource from target.
	 */

	// * query string containing method, resource's URI and protocol version
	$query = "$_SERVER[REQUEST_METHOD] $URI $_SERVER[SERVER_PROTOCOL]";

	// * start header compilation with built query as first line
	$request = array( $query );

	// set hostname
	$request['Host'] = $TARGET['host'];

	// add request header fields available currently
	foreach ( array(
				'HTTP_ACCEPT'            => 'Accept',
				'HTTP_ACCEPT_LANGUAGE'   => 'Accept-Language',
				'HTTP_ACCEPT_CHARSET'    => 'Accept-Charset',
				'HTTP_USER_AGENT'        => 'User-Agent',
				'CONTENT_TYPE'           => 'Content-Type',
				) as $source => $target )
		if ( $_SERVER[$source] )
			$request[$target] = $_SERVER[$source];

	// add previously set cookies
	if ( !is_array( $_SESSION['COOKIES'] ) )
		// initialize session-based local pool of managed cookies
		$_SESSION['COOKIES'] = array();

	$cookies = array();
	foreach ( $_SESSION['COOKIES'] as $key => $cookie )
	{

		if ( isset( $_COOKIE[$key] ) )
			// client provided cookie value
			if ( $passCookies && ( $key != session_name() ) )
				// prefer client-provided cookie valueover any stored on proxy
				$cookie['value'] = strval( $_COOKIE[$key] );

		$cookies[] = $cookie;

	}

	if ( count( $cookies ) )
		$request[] = 'Cookie: ' . compileSendCookies( $cookies );

	// map referer and add as well
	if ( $_SERVER['HTTP_REFERER'] )
		$request['Referer'] = mapToTarget( $_SERVER['HTTP_REFERER'] );

	// support GZIP-encoded data if related function for decoding is available
	if ( is_callable( 'gzdecode' ) )
		$request['Accept-Encoding'] = 'gzip';


	// read request body data
	if ( $_SERVER['REQUEST_METHOD'] == 'POST' )
		$post = file_get_contents( 'php://input' );
	else
		$post = '';

	// add content size to header
	$request['Content-Length'] = strlen( $post );


	// finally compile request data
	foreach ( $request as $key => $value )
		if ( ctype_digit( trim( $key ) ) )
			$request[$key] = "$value\r\n";
		else
			$request[$key] = "$key: $value\r\n";

	$request = implode( '', $request ) . "\r\n$post";




	/*
	 * Send request to target server.
	 */

	// connect to server
	if ( $usePersistentSocket )
		$socket = pfsockopen( $TARGET['host'], $PORT, $ERRNO, $ERRSTR, $TIMEOUT );
	else
		$socket = fsockopen( $TARGET['host'], $PORT, $ERRNO, $ERRSTR, $TIMEOUT );

	if ( !$socket )
		throw new Exception( 'Failed to connect to target!' );

	if ( !stream_set_timeout( $socket, $TIMEOUT, 0 ) )
		throw new Exception( 'Failed to configure link!' );

	// send request
	if ( !fwrite( $socket, $request ) )
		throw new Exception( 'Failed to send request!' );





	/*
	 * Receive response from target server seaparating header from body.
	 */

	// read response
	$header  = $data = $sep = null;
	$expect  = 0;
	$lb      = "\r\n";
	$timeout = 0;
	$dummy   = null;
	$chunked = false;
	$chunkSize = '';

	do
	{

		$reads = array( $socket );
		$s     = stream_select( $reads, $dummy, $dummy, 1, 0 );
		if ( $s === false )
			throw new Exception( 'Failed to wait for incoming data!' );

		if ( !$s )
		{
			if ( ++$timeout == $TIMEOUT )
				throw new Exception( 'Timed out while receiving data!' );
		}
		else
		{

			// relax on waiting for timeout
			$timeout = max( 0, $timeout -1 );

			// get another character
			$ch = fread( $socket, 1 );

			if ( is_null( $data ) )
			{
				//
				// receiving header
				//

				$header .= $ch;

				if ( $ch === $lb[strlen($sep)&1] )
				{

					$sep .= $ch;
					if ( strlen( $sep ) === 4 )
					{
						// met header/data separator

						// --> parse header
						list( $state, $header ) = parseHTTPHeader( trim( $header ) );

						// --> mark to start receiving data on next cycle
						$expect = intval( $header['Content-Length'] );
						$data   = '';

						$chunked = !strcasecmp( $header['Transfer-Encoding'], 'chunked' );

					}
				}
				else
					$sep = '';

			}
			else
			{
				//
				// receiving body
				//

				if ( $chunked )
					// reading in chunks, so checking to read size of chunks
					if ( $expect == 0 )
					{
						// read size of next chunk

						if ( $ch == "\n" )
						{
							if ( $chunkSize !== "\r" )
							{

								$p = substr( $chunkSize, -1 );
								if ( !ctype_xdigit( $p ) && ( $p != "\r" ) )
									throw new Exception( 'Invalid chunk size! (5)' );

								$expect    = intval( hexdec( trim( $chunkSize ) ) );
								$chunkSize = '';

								if ( $expect === 0 )
									break;

								continue;

							}
						}
						else switch ( $ch )
						{

							case "\r" :
								if ( $chunkSize !== '' )
									if ( !ctype_xdigit( trim( $chunkSize ) ) )
										throw new Exception( 'Invalid chunk size! (1)' );
								break;

							case ' ' :
								if ( !ctype_xdigit( trim( $chunkSize ) ) )
									throw new Exception( 'Invalid chunk size! (2)' );
								break;

							default :
								if ( !ctype_xdigit( trim( $ch ) ) )
									throw new Exception( 'Invalid chunk size! (3)' );

								if ( !ctype_xdigit( ltrim( $chunkSize ) ) )
									if ( ( $chunkSize !== '' ) && ( $chunkSize !== "\r\n" ) )
										throw new Exception( 'Invalid chunk size! (4)' );

								break;

						}

						$chunkSize .= $ch;

						continue;

					}


				$data .= $ch;

				$expect--;

			}
		}
	}
	while ( !feof( $socket ) && ( $chunked || is_null( $data ) || ( $expect > 0 ) ) );

	if ( !$usePersistentSocket )
		fclose( $socket );

	if ( is_null( $data ) || ( $expect > 0 ) )
		throw new Exception( 'Incomplete response from server!' );




	/*
	 * Optionally adjust retrieved resource mapping URLs in HTML documents.
	 */

	if ( $data !== '' )
	{

		$ctype = trim( strtok( $header['Content-Type'], ';' ) );
		if ( $ctype == 'text/html' )
		{

			if ( $header['Content-Encoding'] === 'gzip' )
				$data = gzdecode( $data );


			// extract additional information on used charset from ContentType
			$options = array();
			foreach ( explode( ';', strtok( '' ) ) as $option )
			{
				$name = strtolower( trim( strtok( $option, '=' ) ) );
				$value = trim( strtok( '' ) );
				$options[$name] = $value;
			}

			if ( !$options['charset'] )
				$options['charset'] = $defaultCharset;


			// manage to handle trouble on processing pages in UTF-8
			if ( is_callable( 'mb_convert_encoding' ) )
				$data = mb_convert_encoding( $data, 'HTML', $options['charset'] );


			// load HTML document into DOM parser
			$dom = new DOMDocument( '1.0', $options['charset'] );
			$dom->encoding = $options['charset'];
			if ( !@$dom->loadHTML( $data ) )
				throw new Exception( 'Failed to parse HTML!' );


			// list of tags and their attributes to be mapped
			$list = array(
						'A'      => 'HREF',
						'LINK'   => 'HREF',
						'IMG'    => 'SRC',
						'SCRIPT' => 'SRC',
						'STYLE'  => 'SRC',
						'FORM'   => 'ACTION',
						'FRAME'  => 'SRC',
						'EMBED'  => 'SRC',
						'OBJECT' => 'DATA',
						'AREA'   => 'HREF',
						);

			// map optional URLs ...
			foreach ( $list as $tag => $attrs )
			{

				if ( is_null( $lowerNames ) )
				{
					// auto-detect whether names should be lowered or not

					$nodes = $dom->getElementsByTagName( $tag );
					if ( $nodes->length > 0 )
						$lowerNames = false;
					else
					{

						$nodes = $dom->getElementsByTagName( strtolower( $tag));
						if ( $nodes->length > 0 )
							$lowerNames = true;
						else
							continue;

					}
				}

				if ( $lowerNames )
				{
					$tag   = strtolower( $tag );
					$attrs = strtolower( $attrs );
				}

				$attrs = explode( ',', $attrs );

				$nodes = $dom->getElementsByTagName( $tag );
				foreach ( $nodes as $node )
					if ( $node instanceof DOMElement )
						foreach ( $attrs as $attr )
							if ( $node->hasAttribute( $attr ) )
								$node->setAttribute( $attr, mapFromTarget(
												$node->getAttribute( $attr ) ));

			}

			$data = $dom->saveHTML();

			if ( $header['Content-Encoding'] === 'gzip' )
				$data = gzencode( $data );

			$header['Content-Length'] = strlen( $data );

		}

		if ( $chunked )
		{
			// got resource chunked, but won't send it chunked to client
			$header['Content-Length'] = strlen( $data );
			unset( $header['Transfer-Encoding'] );
		}

	}



	/*
	 * Add record to local cache.
	 */

	if ( $data && $cacheHash && is_dir( $cacheDir ) && is_writable( $cacheDir ))
	{

		if ( is_string( $useCacheMIME ) )
		{

			$pattern = trim( $useCacheMIME );
			$negate  = ( $pattern[0] == '!' );
			if ( $negate )
				$pattern = substr( $pattern, 1 );

			$useCacheMIME = preg_match( $pattern, $ctype ) ^ $negate;

		}
		else
			$useCacheMIME = false;

		if ( $useCacheMIME )
			file_put_contents( "$cacheDir/$cacheHash",
							   serialize( array( $state, $header, $data ) ) );

	}



	/*
	 * pass received data to client
	 */

	sendResponse( $state, $header, $data );

}
catch ( Exception $e )
{
	echo 'toxA PHP Proxy: ' . $e->getMessage() . " / $ERRNO - $ERRSTR";
}



/*
 * Append some information for debugging.
 */

if ( $debug && ( $ctype == 'text/html' ) )
{
?>
<div class="debug">
 <h2>Proxy Cookie Pool</h2>
 <pre><?php var_dump( $_SESSION['COOKIES'] ) ?></pre>
 <h2>Request To Target</h2>
 <pre><?php echo $request ?></pre>
 <h2>Response Header From Target</h2>
 <pre><?php
echo "$state[raw]\n";
foreach ( $header as $name => $row )
{
	if ( !is_array( $row ) )
		$row = array( $row );

	foreach ( $row as $value )
		echo "$name: $value\n";

}
?></pre>
</div>
<?php
}

exit( 0 );






/*
 * --------- methods supporting code above ---------
 */


function mapFromTarget( $in )
{
	global $PROXY, $TARGET;

	return mapPath( $in, $TARGET, $PROXY );
}

function mapToTarget( $in )
{
	global $PROXY, $TARGET;

	return mapPath( $in, $PROXY, $TARGET );
}

function mapPath( $in, $from, $to )
{

	$temp = $orig = @parse_url( $in );
	if ( $temp )
	{

		if ( $temp['scheme'] )
		{
			if ( $temp['scheme'] !== $from['scheme'] )
				return $in;
			$temp['scheme'] = $to['scheme'];
		}

		if ( $temp['host'] )
		{
			if ( $temp['host'] !== $from['host'] )
				return $in;
			$temp['host'] = $to['host'];
		}


		$uri  = $temp['path'];

		$pos  = strlen( $from['path'] );
		if ( ( $pos === 0 ) || ( substr( $uri, 0, $pos ) === $from['path'] ) )
		{

			$uri = substr( $uri, $pos );
			if ( ( $uri === '' ) || ( $uri[0] === '/' ) )
				$temp['path'] = $to['path'] . $uri;

		}


		if ( $temp != $orig )
		{

			$out = '';
			if ( $temp['scheme'] )
				$out .= $temp['scheme'] . ':';

			if ( $temp['pass'] )
				$temp['user'] .= ':' . $temp['pass'];

			if ( $temp['user'] )
				$temp['host'] = $temp['user'] . '@' . $temp['host'];

			if ( $temp['host'] )
				$out .= '//' . $temp['host'];

			if ( $temp['path'] )
				$out .= $temp['path'];

			if ( $temp['query'] )
				$out .= '?' . $temp['query'];

			if ( $temp['fragment'] )
				$out .= '#' . $temp['fragment'];


			return $out;

		}
	}


	return $in;

}


function parseHTTPHeader( $data )
{
	global $passCookies;


	$state  = null;
	$header = array();


	/*
	 * Parse HTTP header line by line collecting information in hash array.
	 */

	foreach ( explode( "\r\n", $data ) as $line )
		if ( is_null( $state ) )
		{
			// read HTTP status first

			$line = trim( $line );

			$protocol = strtok( $line, ' ' );
			$code     = strtok( ' ' );
			$text     = strtok( '' );

			$state = array(
							'raw'      => $line,
							'protocol' => $protocol,
							'code'     => $code,
							'text'     => $text,
							);

		}
		else
		{
			// expect more HTTP header lines

			if ( ctype_space( $line[0] ) )
			{
				// folded line

				if ( !$name )
					throw new Exception( 'Invalid folded line!' );

				if ( is_array( $header[$name] ) )
				{
					$line = array_pop( $header[$name] );
					array_push( $header[$name], $line . $value );
				}
				else
					$header[$name] .= $value;

			}
			else
			{

				$name  = strtok( $line, ':' );
				$value = trim( strtok( '' ) );

				// normalize name
				$name = str_replace( ' ', '-', ucwords( str_replace( '-', ' ', $name ) ) );


				if ( !isset( $header[$name] ) )
					$header[$name] = $value;
				else
				{

					if ( is_string( $header[$name] ) )
						$header[$name] = array( $header[$name] );

					$header[$name][] = $value;

				}
			}
		}



	/*
	 * Fix pathes.
	 */

	if ( $header['Location'] )
		$header['Location'] = mapFromTarget( $header['Location'] );



	/*
	 * Update proxy-local cookies pool.
	 */

	if ( $header['Set-Cookie'] )
	{

		if ( !is_array( $header['Set-Cookie'] ) )
			$header['Set-Cookie'] = array( $header['Set-Cookie'] );

		foreach ( $header['Set-Cookie'] as $key => $cookie )
		{

			$cookie = parseCookie( $cookie );
			if ( $cookie )
			{

				// store each cookie in proxy-local cookie pool
				$_SESSION['COOKIES'][$cookie['name']] = $cookie;

				if ( $passCookies && ( $cookie['name'] !== session_name() ) )
				{
					// fix cookie in header to be passed back to client

					if ( $cookie['options']['Path'] )
						$cookie['options']['Path'] = mapFromTarget(
												$cookie['options']['Path']);

					$header['Set-Cookie'][$key] = compileSetCookie( $cookie );

				}
				else
					unset( $header['Set-Cookie'][$key] );

			}
		}
	}


	return array( $state, $header );

}


function parseCookie( $in )
{

	$out = array(
				'name'    => '',
				'value'   => '',
				'options' => array(),
				);


	while ( trim( $in ) !== '' )
	{

		$length = strcspn( $in, '=' );
		if ( !$length )
			break;

		$name  = trim( substr( $in, 0, $length ) );
		$in    = trim( substr( $in, $length + 1 ) );

		$value = stripcslashes( getString( $in, ';', true ) );

		$in    = trim( $in );
		if ( $in[0] == ';' )
			$in = trim( substr( $in, 1 ) );


		if ( $out['name'] === '' )
		{
			// store cookie name and value

			$out['name']  = $name;
			$out['value'] = $value;

		}
		else
		{
			// add option

			if ( $name[0] == '$' )
				$name = substr( $name, 1 );

			$name = ucfirst( $name );

			$out['options'][$name] = $value;

		}
	}


	return ( $out['name'] !== '' ) ? $out : false;

}


function compileSetCookie( $in )
{

	if ( !$in['name'] )
		return '';


	if ( strpos( $in['value'], '"' ) !== false )
		$in['value'] = '"' . addcslashes( $in['value'], "\"\n\r" ) . '"';
	else
		$in['value'] = addcslashes( $in['value'], "\n\r" );

	$out = $in['name'] . '=' . $in['value'];


	foreach ( $in['options'] as $name => $value )
	{

		if ( strpos( $value, '"' ) !== false )
			$value = '"' . addcslashes( $value, "\"\n\r" ) . '"';
		else
			$value = addcslashes( $value, "\n\r" );


		$in['options'][$name] = $name . '=' . $value . ';';

	}

	if ( count( $in['options'] ) )
		$out .= '; ' . implode( ' ', $in['options'] );


	return $out;

}


function compileSendCookies( $in )
{

	if ( !is_array( $in ) )
		return null;


	$out = array();

	foreach ( $in as $cookie )
	{

		if ( !$cookie['name'] )
			continue;

		if ( strpos( $cookie['value'], '"' ) !== false )
			$cookie['value'] = '"' . addcslashes( $cookie['value'], "\"\n\r" ) . '"';
		else
			$cookie['value'] = addcslashes( $cookie['value'], "\n\r" );

		$out[] = $cookie['name'] . '=' . $cookie['value'];

	}

	if ( count( $out ) )
		array_unshift( $out, '$Version="1"' );


	return implode( '; ', $out );

}


function getString( &$in, $sep, $only = false )
{

	$in = ltrim( $in );

	if ( $in[0] == '"' )
	{

		$opener = '"';
		$pos    = 1;

		do
		{

			$length = strcspn( $in, '"', $pos );
			if ( $length == 0 )
				break;

			for ( $i = 1; $i < $length; $i++ )
				if ( $in[$pos+$length-$i] != "\\" )
					break;

			$pos += $length;

			if ( $i & 1 )
				// double-quote is preceeded by even number of backslashes
				// --> ignore this double-quote
				$pos++;
			else
				break;

		}
		while ( $length <= strlen( $in ) );


		$out = substr( $in, 1, $pos - 2 );
		$in  = substr( $in, $pos );

	}
	else
	{

		$length = strcspn( $in, $only ? $sep : " \r\n\t\f$sep" );
		$out    = substr( $in, 0, $length );
		$in     = substr( $in, $length );

	}


	return $out;

}


function sendResponse( $state, $header, $data )
{
	global $PROXY, $version;



	/*
	 * pass one last validation of data to be sent to client
	 */

	// either don't include any Set-Cookie-Request or include proxy's PHPSESSID
	if ( empty( $header['Set-Cookie'] ) )
		unset( $header['Set-Cookie'] );
	else
		$header['Set-Cookie'][] = compileSetCookie( array(
										'name'    => session_name(),
										'value'   => session_id(),
										'options' => array(
												'Path'   => $PROXY['path'],
												'Domain' => $PROXY['host'],
												),
										) );



	header( $state['raw'] );

	foreach ( $header as $name => $value )
		if ( $debug && ( $name == 'Content-Length' ) )
			continue;
		else if ( is_array( $value ) )
			foreach ( array_values( $value ) as $k => $v )
				header( "$name: $v", !$k );
		else
			header( "$name: $value" );

	header( 'X-Proxy-Version: toxA PHP-Proxy ' . $version );


	echo $data;

}


?>